import { Item } from './item.js';
import { CollectionPersistent } from './collection.persistent.js';
import { defineConfig, generateId, normalizeArray, isValidObject, isFunction, removeProperties, copy } from '@agile-ts/utils';
import { LogCodeManager } from '../logCodeManager.js';
import { Selector } from './selector.js';
import { Group, TrackedChangeMethod } from './group/index.js';
import { ComputedTracker } from '../computed/computed.tracker.js';

var __defProp = Object.defineProperty;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
typeof require !== "undefined" ? require : (x) => {
  throw new Error('Dynamic require of "' + x + '" is not supported');
};
class Collection {
  constructor(agileInstance, config = {}) {
    this.size = 0;
    this.data = {};
    this.isPersisted = false;
    this.groups = {};
    this.selectors = {};
    this.isInstantiated = false;
    this.isCollection = true;
    this.agileInstance = () => agileInstance;
    let _config = typeof config === "function" ? config(this) : config;
    _config = defineConfig(_config, {
      primaryKey: "id",
      groups: {},
      selectors: {},
      defaultGroupKey: "default"
    });
    this._key = _config.key;
    this.config = {
      defaultGroupKey: _config.defaultGroupKey,
      primaryKey: _config.primaryKey
    };
    this.initialConfig = _config;
    this.initGroups(_config.groups);
    this.initSelectors(_config.selectors);
    this.isInstantiated = true;
    if (_config.initialData)
      this.collect(_config.initialData);
    for (const key in this.selectors)
      this.selectors[key].reselect();
  }
  set key(value) {
    this.setKey(value);
  }
  get key() {
    return this._key;
  }
  setKey(value) {
    var _a, _b;
    const oldKey = this._key;
    this._key = value;
    if (value != null && ((_a = this.persistent) == null ? void 0 : _a._key) === oldKey)
      (_b = this.persistent) == null ? void 0 : _b.setKey(value);
    return this;
  }
  Group(initialItems, config = {}) {
    var _a;
    if (this.isInstantiated) {
      const key = (_a = config.key) != null ? _a : generateId();
      LogCodeManager.log("1B:02:00");
      return this.createGroup(key, initialItems);
    }
    return new Group(this, initialItems, config);
  }
  Selector(initialKey, config = {}) {
    var _a;
    if (this.isInstantiated) {
      const key = (_a = config.key) != null ? _a : generateId();
      LogCodeManager.log("1B:02:01");
      return this.createSelector(key, initialKey);
    }
    return new Selector(this, initialKey, config);
  }
  initGroups(groups) {
    if (!groups)
      return;
    let groupsObject = {};
    if (Array.isArray(groups)) {
      groups.forEach((groupKey) => {
        groupsObject[groupKey] = new Group(this, [], {
          key: groupKey
        });
      });
    } else
      groupsObject = groups;
    groupsObject[this.config.defaultGroupKey] = new Group(this, [], {
      key: this.config.defaultGroupKey
    });
    for (const key in groupsObject)
      if (groupsObject[key]._key == null)
        groupsObject[key].setKey(key);
    this.groups = groupsObject;
  }
  initSelectors(selectors) {
    if (!selectors)
      return;
    let selectorsObject = {};
    if (Array.isArray(selectors)) {
      selectors.forEach((selectorKey) => {
        selectorsObject[selectorKey] = new Selector(this, selectorKey, {
          key: selectorKey
        });
      });
    } else
      selectorsObject = selectors;
    for (const key in selectorsObject)
      if (selectorsObject[key]._key == null)
        selectorsObject[key].setKey(key);
    this.selectors = selectorsObject;
  }
  collect(data, groupKeys, config = {}) {
    const _data = normalizeArray(data);
    const _groupKeys = normalizeArray(groupKeys);
    const defaultGroupKey = this.config.defaultGroupKey;
    const primaryKey = this.config.primaryKey;
    config = defineConfig(config, {
      method: "push",
      background: false,
      patch: false,
      select: false
    });
    if (!_groupKeys.includes(defaultGroupKey))
      _groupKeys.push(defaultGroupKey);
    _groupKeys.forEach((key) => this.groups[key] == null && this.createGroup(key));
    _data.forEach((data2, index) => {
      let itemKey;
      let success = false;
      if (data2 instanceof Item) {
        success = this.assignItem(data2, {
          background: config.background
        });
        itemKey = data2._key;
      } else {
        success = this.assignData(data2, {
          patch: config.patch,
          background: config.background
        });
        itemKey = data2[primaryKey];
      }
      if (success) {
        _groupKeys.forEach((groupKey) => {
          var _a;
          (_a = this.getGroup(groupKey)) == null ? void 0 : _a.add(itemKey, {
            method: config.method,
            background: config.background
          });
        });
        if (config.select)
          this.createSelector(itemKey, itemKey);
      }
      if (config.forEachItem)
        config.forEachItem(data2, itemKey, success, index);
    });
    return this;
  }
  update(itemKey, changes, config = {}) {
    const item = this.getItem(itemKey, { notExisting: true });
    const primaryKey = this.config.primaryKey;
    config = defineConfig(config, {
      patch: true,
      background: false
    });
    if (item == null) {
      LogCodeManager.log("1B:03:00", [itemKey, this._key]);
      return void 0;
    }
    if (!isValidObject(changes)) {
      LogCodeManager.log("1B:03:01", [itemKey, this._key]);
      return void 0;
    }
    const oldItemKey = item._value[primaryKey];
    const newItemKey = changes[primaryKey] || oldItemKey;
    if (oldItemKey !== newItemKey)
      this.updateItemKey(oldItemKey, newItemKey, {
        background: config.background
      });
    if (config.patch) {
      if (changes[primaryKey])
        delete changes[primaryKey];
      let patchConfig = typeof config.patch === "object" ? config.patch : {};
      patchConfig = __spreadValues({
        addNewProperties: true
      }, patchConfig);
      item.patch(changes, {
        background: config.background,
        addNewProperties: patchConfig.addNewProperties
      });
    } else {
      if (changes[this.config.primaryKey] !== itemKey) {
        changes[this.config.primaryKey] = itemKey;
        LogCodeManager.log("1B:02:02", [], changes);
      }
      item.set(changes, {
        background: config.background
      });
    }
    return item;
  }
  createGroup(groupKey, initialItems = []) {
    let group = this.getGroup(groupKey, { notExisting: true });
    if (!this.isInstantiated)
      LogCodeManager.log("1B:02:03");
    if (group != null) {
      if (!group.isPlaceholder) {
        LogCodeManager.log("1B:03:02", [groupKey]);
        return group;
      }
      group.set(initialItems, { overwrite: true });
      return group;
    }
    group = new Group(this, initialItems, { key: groupKey });
    this.groups[groupKey] = group;
    return group;
  }
  hasGroup(groupKey, config = {}) {
    return !!this.getGroup(groupKey, config);
  }
  getGroup(groupKey, config = {}) {
    config = defineConfig(config, {
      notExisting: false
    });
    const group = groupKey ? this.groups[groupKey] : void 0;
    if (group == null || !config.notExisting && !group.exists)
      return void 0;
    ComputedTracker.tracked(group.observers["value"]);
    return group;
  }
  getDefaultGroup() {
    return this.getGroup(this.config.defaultGroupKey);
  }
  getGroupWithReference(groupKey) {
    let group = this.getGroup(groupKey, { notExisting: true });
    if (group == null) {
      group = new Group(this, [], {
        key: groupKey,
        isPlaceholder: true
      });
      this.groups[groupKey] = group;
    }
    ComputedTracker.tracked(group.observers["value"]);
    return group;
  }
  removeGroup(groupKey) {
    if (this.groups[groupKey] != null)
      delete this.groups[groupKey];
    return this;
  }
  getGroupCount() {
    let size = 0;
    Object.keys(this.groups).map(() => size++);
    return size;
  }
  createSelector(selectorKey, itemKey) {
    let selector = this.getSelector(selectorKey, { notExisting: true });
    if (!this.isInstantiated)
      LogCodeManager.log("1B:02:04");
    if (selector != null) {
      if (!selector.isPlaceholder) {
        LogCodeManager.log("1B:03:03", [selectorKey]);
        return selector;
      }
      selector.select(itemKey, { overwrite: true });
      return selector;
    }
    selector = new Selector(this, itemKey, {
      key: selectorKey
    });
    this.selectors[selectorKey] = selector;
    return selector;
  }
  select(itemKey) {
    return this.createSelector(itemKey, itemKey);
  }
  hasSelector(selectorKey, config = {}) {
    return !!this.getSelector(selectorKey, config);
  }
  getSelector(selectorKey, config = {}) {
    config = defineConfig(config, {
      notExisting: false
    });
    const selector = selectorKey ? this.selectors[selectorKey] : void 0;
    if (selector == null || !config.notExisting && !selector.exists)
      return void 0;
    ComputedTracker.tracked(selector.observers["value"]);
    return selector;
  }
  getSelectorWithReference(selectorKey) {
    let selector = this.getSelector(selectorKey, { notExisting: true });
    if (selector == null) {
      selector = new Selector(this, null, {
        key: selectorKey,
        isPlaceholder: true
      });
      this.selectors[selectorKey] = selector;
    }
    ComputedTracker.tracked(selector.observers["value"]);
    return selector;
  }
  removeSelector(selectorKey) {
    if (this.selectors[selectorKey] != null) {
      this.selectors[selectorKey].unselect();
      delete this.selectors[selectorKey];
    }
    return this;
  }
  getSelectorCount() {
    let size = 0;
    Object.keys(this.selectors).map(() => size++);
    return size;
  }
  hasItem(itemKey, config = {}) {
    return !!this.getItem(itemKey, config);
  }
  getItem(itemKey, config = {}) {
    config = defineConfig(config, {
      notExisting: false
    });
    const item = itemKey != null ? this.data[itemKey] : void 0;
    if (item == null || !config.notExisting && !item.exists)
      return void 0;
    ComputedTracker.tracked(item.observers["value"]);
    return item;
  }
  getItemWithReference(itemKey) {
    let item = this.getItem(itemKey, { notExisting: true });
    if (item == null)
      item = this.createPlaceholderItem(itemKey, true);
    ComputedTracker.tracked(item.observers["value"]);
    return item;
  }
  createPlaceholderItem(itemKey, addToCollection = false) {
    const item = new Item(this, {
      [this.config.primaryKey]: itemKey,
      dummy: "item"
    }, { isPlaceholder: true });
    if (addToCollection && !Object.prototype.hasOwnProperty.call(this.data, itemKey))
      this.data[itemKey] = item;
    ComputedTracker.tracked(item.observers["value"]);
    return item;
  }
  getItemValue(itemKey, config = {}) {
    const item = this.getItem(itemKey, config);
    if (item == null)
      return void 0;
    return item.value;
  }
  getAllItems(config = {}) {
    config = defineConfig(config, {
      notExisting: false
    });
    const defaultGroup = this.getDefaultGroup();
    let items = [];
    if (config.notExisting) {
      for (const key in this.data)
        items.push(this.data[key]);
    } else {
      items = (defaultGroup == null ? void 0 : defaultGroup.getItems()) || [];
    }
    return items;
  }
  getAllItemValues(config = {}) {
    const items = this.getAllItems(config);
    return items.map((item) => item.value);
  }
  persist(keyOrConfig = {}, config = {}) {
    let _config;
    let key;
    if (isValidObject(keyOrConfig)) {
      _config = keyOrConfig;
      key = this._key;
    } else {
      _config = config || {};
      key = keyOrConfig;
    }
    _config = defineConfig(_config, {
      loadValue: true,
      storageKeys: [],
      defaultStorageKey: null
    });
    if (this.persistent != null && this.isPersisted)
      return this;
    this.persistent = new CollectionPersistent(this, {
      loadValue: _config.loadValue,
      storageKeys: _config.storageKeys,
      key,
      defaultStorageKey: _config.defaultStorageKey
    });
    return this;
  }
  onLoad(callback) {
    if (!this.persistent)
      return this;
    if (!isFunction(callback)) {
      LogCodeManager.log("00:03:01", ["OnLoad Callback", "function"]);
      return this;
    }
    this.persistent.onLoad = callback;
    if (this.isPersisted)
      callback(true);
    return this;
  }
  reset() {
    var _a, _b;
    this.data = {};
    this.size = 0;
    for (const key in this.groups)
      (_a = this.getGroup(key)) == null ? void 0 : _a.reset();
    for (const key in this.selectors)
      (_b = this.getSelector(key)) == null ? void 0 : _b.reset();
    return this;
  }
  put(itemKeys, groupKeys, config = {}) {
    const _itemKeys = normalizeArray(itemKeys);
    const _groupKeys = normalizeArray(groupKeys);
    _groupKeys.forEach((groupKey) => {
      var _a;
      (_a = this.getGroup(groupKey)) == null ? void 0 : _a.add(_itemKeys, config);
    });
    return this;
  }
  move(itemKeys, oldGroupKey, newGroupKey, config = {}) {
    var _a, _b;
    const _itemKeys = normalizeArray(itemKeys);
    (_a = this.getGroup(oldGroupKey)) == null ? void 0 : _a.remove(_itemKeys, removeProperties(config, ["method", "overwrite"]));
    (_b = this.getGroup(newGroupKey)) == null ? void 0 : _b.add(_itemKeys, config);
    return this;
  }
  updateItemKey(oldItemKey, newItemKey, config = {}) {
    var _a;
    const item = this.getItem(oldItemKey, { notExisting: true });
    config = defineConfig(config, {
      background: false
    });
    if (item == null || oldItemKey === newItemKey)
      return false;
    if (this.hasItem(newItemKey)) {
      LogCodeManager.log("1B:03:04", [oldItemKey, newItemKey, this._key]);
      return false;
    }
    delete this.data[oldItemKey];
    this.data[newItemKey] = item;
    item.setKey(newItemKey, {
      background: config.background
    });
    if (item.persistent != null && item.persistent._key === CollectionPersistent.getItemStorageKey(oldItemKey, this._key))
      (_a = item.persistent) == null ? void 0 : _a.setKey(CollectionPersistent.getItemStorageKey(newItemKey, this._key));
    for (const groupKey in this.groups) {
      const group = this.getGroup(groupKey, { notExisting: true });
      if (group == null || !group.has(oldItemKey))
        continue;
      group.replace(oldItemKey, newItemKey, { background: config.background });
    }
    for (const selectorKey in this.selectors) {
      const selector = this.getSelector(selectorKey, { notExisting: true });
      if (selector == null)
        continue;
      if (selector.hasSelected(newItemKey, false)) {
        selector.reselect({
          force: true,
          background: config.background
        });
      }
      if (selector.hasSelected(oldItemKey, false))
        selector.select(newItemKey, {
          background: config.background
        });
    }
    return true;
  }
  getGroupKeysThatHaveItemKey(itemKey) {
    const groupKeys = [];
    for (const groupKey in this.groups) {
      const group = this.groups[groupKey];
      if (group == null ? void 0 : group.has(itemKey))
        groupKeys.push(groupKey);
    }
    return groupKeys;
  }
  remove(itemKeys) {
    return {
      fromGroups: (groups) => this.removeFromGroups(itemKeys, groups),
      everywhere: (config) => this.removeItems(itemKeys, config || {})
    };
  }
  removeFromGroups(itemKeys, groupKeys) {
    const _itemKeys = normalizeArray(itemKeys);
    const _groupKeys = normalizeArray(groupKeys);
    _itemKeys.forEach((itemKey) => {
      let removedFromGroupsCount = 0;
      _groupKeys.forEach((groupKey) => {
        const group = this.getGroup(groupKey, { notExisting: true });
        if (!(group == null ? void 0 : group.has(itemKey)))
          return;
        group.remove(itemKey);
        removedFromGroupsCount++;
      });
      if (removedFromGroupsCount >= this.getGroupKeysThatHaveItemKey(itemKey).length)
        this.removeItems(itemKey);
    });
    return this;
  }
  removeItems(itemKeys, config = {}) {
    config = defineConfig(config, {
      notExisting: false,
      removeSelector: false
    });
    const _itemKeys = normalizeArray(itemKeys);
    _itemKeys.forEach((itemKey) => {
      var _a, _b;
      const item = this.getItem(itemKey, { notExisting: config.notExisting });
      if (item == null)
        return;
      const wasPlaceholder = item.isPlaceholder;
      for (const groupKey in this.groups) {
        const group = this.getGroup(groupKey, { notExisting: true });
        if (group == null ? void 0 : group.has(itemKey))
          group == null ? void 0 : group.remove(itemKey);
      }
      (_a = item.persistent) == null ? void 0 : _a.removePersistedValue();
      delete this.data[itemKey];
      for (const selectorKey in this.selectors) {
        const selector = this.getSelector(selectorKey, { notExisting: true });
        if (selector != null && selector.hasSelected(itemKey, false)) {
          if (config.removeSelector) {
            this.removeSelector((_b = selector._key) != null ? _b : "unknown");
          } else {
            selector.reselect({ force: true });
          }
        }
      }
      if (!wasPlaceholder)
        this.size--;
    });
    return this;
  }
  assignData(data, config = {}) {
    config = defineConfig(config, {
      patch: false,
      background: false
    });
    const _data = copy(data);
    const primaryKey = this.config.primaryKey;
    if (!isValidObject(_data)) {
      LogCodeManager.log("1B:03:05", [this._key]);
      return false;
    }
    if (!Object.prototype.hasOwnProperty.call(_data, primaryKey)) {
      LogCodeManager.log("1B:02:05", [this._key, primaryKey]);
      _data[primaryKey] = generateId();
    }
    const itemKey = _data[primaryKey];
    const item = this.getItem(itemKey, { notExisting: true });
    const wasPlaceholder = (item == null ? void 0 : item.isPlaceholder) || false;
    if (item != null) {
      if (config.patch) {
        item.patch(_data, { background: config.background });
      } else {
        item.set(_data, { background: config.background });
      }
    } else {
      this.assignItem(new Item(this, _data), {
        background: config.background
      });
    }
    if (wasPlaceholder)
      this.size++;
    return true;
  }
  assignItem(item, config = {}) {
    config = defineConfig(config, {
      overwrite: false,
      background: false,
      rebuildGroups: true
    });
    const primaryKey = this.config.primaryKey;
    let itemKey = item._value[primaryKey];
    let increaseCollectionSize = true;
    if (!Object.prototype.hasOwnProperty.call(item._value, primaryKey)) {
      LogCodeManager.log("1B:02:05", [this._key, primaryKey]);
      itemKey = generateId();
      item.patch({ [this.config.primaryKey]: itemKey }, { background: config.background });
      item._key = itemKey;
    }
    if (item.collection() !== this) {
      LogCodeManager.log("1B:03:06", [this._key, item.collection()._key]);
      return false;
    }
    if (this.getItem(itemKey, { notExisting: true }) != null) {
      if (!config.overwrite) {
        this.assignData(item._value);
        return true;
      } else
        increaseCollectionSize = false;
    }
    this.data[itemKey] = item;
    if (config.rebuildGroups)
      this.rebuildGroupsThatIncludeItemKey(itemKey, {
        background: config.background
      });
    if (increaseCollectionSize)
      this.size++;
    return true;
  }
  rebuildGroupsThatIncludeItemKey(itemKey, config = {}) {
    for (const groupKey of Object.keys(this.groups)) {
      const group = this.getGroup(groupKey);
      if (group != null && group.has(itemKey)) {
        const index = group._preciseItemKeys.findIndex((ik) => itemKey === ik);
        if (index !== -1) {
          group.rebuild([
            {
              key: itemKey,
              index,
              method: TrackedChangeMethod.UPDATE
            }
          ], config);
        } else {
          const indexOfBeforeItemKey = group.nextStateValue.findIndex((ik) => itemKey === ik) - 1;
          group.rebuild([
            {
              key: itemKey,
              index: indexOfBeforeItemKey >= 0 ? group._preciseItemKeys.findIndex((ik) => group.nextStateValue[indexOfBeforeItemKey] === ik) + 1 : 0,
              method: TrackedChangeMethod.ADD
            }
          ], config);
        }
      }
    }
  }
}

export { Collection };
