import { defineConfig } from '@agile-ts/utils';
import { LogCodeManager } from '../logCodeManager.js';
import { Persistent } from '../storages/persistent.js';
import { getStorageManager } from '../storages/shared.js';

typeof require !== "undefined" ? require : (x) => {
  throw new Error('Dynamic require of "' + x + '" is not supported');
};
var __async = (__this, __arguments, generator) => {
  return new Promise((resolve, reject) => {
    var fulfilled = (value) => {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    };
    var rejected = (value) => {
      try {
        step(generator.throw(value));
      } catch (e) {
        reject(e);
      }
    };
    var step = (x) => x.done ? resolve(x.value) : Promise.resolve(x.value).then(fulfilled, rejected);
    step((generator = generator.apply(__this, __arguments)).next());
  });
};
const _CollectionPersistent = class extends Persistent {
  constructor(collection, config = {}) {
    super(collection.agileInstance(), {
      loadValue: false
    });
    config = defineConfig(config, {
      loadValue: true,
      storageKeys: [],
      defaultStorageKey: null
    });
    this.collection = () => collection;
    this.instantiatePersistent({
      key: config.key,
      storageKeys: config.storageKeys,
      defaultStorageKey: config.defaultStorageKey
    });
    if (this.ready && config.loadValue)
      this.initialLoading();
  }
  initialLoading() {
    var __super = (key) => super[key];
    return __async(this, null, function* () {
      __super("initialLoading").call(this).then(() => {
        this.collection().isPersisted = true;
      });
    });
  }
  loadPersistedValue(storageItemKey) {
    return __async(this, null, function* () {
      var _a;
      if (!this.ready)
        return false;
      const _storageItemKey = storageItemKey != null ? storageItemKey : this._key;
      const isPersisted = yield (_a = getStorageManager()) == null ? void 0 : _a.get(_storageItemKey, this.config.defaultStorageKey);
      if (!isPersisted)
        return false;
      const loadValuesIntoCollection = () => __async(this, null, function* () {
        var _a2, _b, _c;
        const defaultGroup = this.collection().getDefaultGroup();
        if (defaultGroup == null)
          return false;
        const defaultGroupStorageKey = _CollectionPersistent.getGroupStorageKey(defaultGroup._key, _storageItemKey);
        defaultGroup.loadedInitialValue = false;
        defaultGroup.persist(defaultGroupStorageKey, {
          loadValue: false,
          defaultStorageKey: this.config.defaultStorageKey || void 0,
          storageKeys: this.storageKeys,
          followCollectionPersistKeyPattern: false
        });
        if ((_a2 = defaultGroup.persistent) == null ? void 0 : _a2.ready)
          yield defaultGroup.persistent.initialLoading();
        for (const itemKey of defaultGroup._value) {
          const item = this.collection().getItem(itemKey);
          const itemStorageKey = _CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
          if (item != null) {
            item.persist(itemStorageKey, {
              loadValue: false,
              defaultStorageKey: this.config.defaultStorageKey || void 0,
              storageKeys: this.storageKeys,
              followCollectionPersistKeyPattern: false
            });
            if ((_b = item == null ? void 0 : item.persistent) == null ? void 0 : _b.ready)
              yield item.persistent.initialLoading();
          } else {
            const placeholderItem = this.collection().getItemWithReference(itemKey);
            placeholderItem == null ? void 0 : placeholderItem.persist(itemStorageKey, {
              loadValue: false,
              defaultStorageKey: this.config.defaultStorageKey,
              storageKeys: this.storageKeys,
              followCollectionPersistKeyPattern: false
            });
            if ((_c = placeholderItem == null ? void 0 : placeholderItem.persistent) == null ? void 0 : _c.ready) {
              const loadedPersistedValueIntoItem = yield placeholderItem.persistent.loadPersistedValue();
              if (loadedPersistedValueIntoItem) {
                this.collection().assignItem(placeholderItem, {
                  overwrite: true,
                  rebuildGroups: false
                });
                placeholderItem.isPersisted = true;
                this.collection().size += 1;
              }
            }
          }
        }
        defaultGroup.loadedInitialValue = true;
        return true;
      });
      const success = yield loadValuesIntoCollection();
      if (success)
        this.setupSideEffects(_storageItemKey);
      return success;
    });
  }
  persistValue(storageItemKey) {
    return __async(this, null, function* () {
      var _a;
      if (!this.ready)
        return false;
      const _storageItemKey = storageItemKey != null ? storageItemKey : this._key;
      const defaultGroup = this.collection().getDefaultGroup();
      if (defaultGroup == null)
        return false;
      const defaultGroupStorageKey = _CollectionPersistent.getGroupStorageKey(defaultGroup._key, _storageItemKey);
      (_a = getStorageManager()) == null ? void 0 : _a.set(_storageItemKey, true, this.storageKeys);
      defaultGroup.persist(defaultGroupStorageKey, {
        defaultStorageKey: this.config.defaultStorageKey || void 0,
        storageKeys: this.storageKeys,
        followCollectionPersistKeyPattern: false
      });
      for (const itemKey of defaultGroup._value) {
        const item = this.collection().getItem(itemKey);
        const itemStorageKey = _CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
        item == null ? void 0 : item.persist(itemStorageKey, {
          defaultStorageKey: this.config.defaultStorageKey || void 0,
          storageKeys: this.storageKeys,
          followCollectionPersistKeyPattern: false
        });
      }
      this.setupSideEffects(_storageItemKey);
      this.isPersisted = true;
      return true;
    });
  }
  setupSideEffects(storageItemKey) {
    const _storageItemKey = storageItemKey != null ? storageItemKey : this._key;
    const defaultGroup = this.collection().getDefaultGroup();
    if (defaultGroup == null)
      return;
    defaultGroup.addSideEffect(_CollectionPersistent.defaultGroupSideEffectKey, (instance) => this.rebuildStorageSideEffect(instance, _storageItemKey), { weight: 0 });
  }
  removePersistedValue(storageItemKey) {
    return __async(this, null, function* () {
      var _a, _b, _c;
      if (!this.ready)
        return false;
      const _storageItemKey = storageItemKey != null ? storageItemKey : this._key;
      const defaultGroup = this.collection().getDefaultGroup();
      if (!defaultGroup)
        return false;
      const defaultGroupStorageKey = _CollectionPersistent.getGroupStorageKey(defaultGroup._key, _storageItemKey);
      (_a = getStorageManager()) == null ? void 0 : _a.remove(_storageItemKey, this.storageKeys);
      (_b = defaultGroup.persistent) == null ? void 0 : _b.removePersistedValue(defaultGroupStorageKey);
      defaultGroup.removeSideEffect(_CollectionPersistent.defaultGroupSideEffectKey);
      for (const itemKey of defaultGroup._value) {
        const item = this.collection().getItem(itemKey);
        const itemStorageKey = _CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
        (_c = item == null ? void 0 : item.persistent) == null ? void 0 : _c.removePersistedValue(itemStorageKey);
      }
      this.isPersisted = false;
      return true;
    });
  }
  formatKey(key) {
    if (key == null && this.collection()._key)
      return this.collection()._key;
    if (key == null)
      return;
    if (this.collection()._key == null)
      this.collection()._key = key;
    return key;
  }
  rebuildStorageSideEffect(group, storageItemKey) {
    var _a;
    const collection = group.collection();
    const _storageItemKey = storageItemKey || ((_a = collection.persistent) == null ? void 0 : _a._key);
    if (group.previousStateValue.length === group._value.length)
      return;
    const addedKeys = group._value.filter((key) => !group.previousStateValue.includes(key));
    const removedKeys = group.previousStateValue.filter((key) => !group._value.includes(key));
    addedKeys.forEach((itemKey) => {
      const item = collection.getItem(itemKey);
      const itemStorageKey = _CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
      if (item != null && !item.isPersisted)
        item.persist(itemStorageKey, {
          defaultStorageKey: this.config.defaultStorageKey || void 0,
          storageKeys: this.storageKeys,
          followCollectionPersistKeyPattern: false
        });
    });
    removedKeys.forEach((itemKey) => {
      var _a2;
      const item = collection.getItem(itemKey);
      const itemStorageKey = _CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
      if (item != null && item.isPersisted)
        (_a2 = item.persistent) == null ? void 0 : _a2.removePersistedValue(itemStorageKey);
    });
  }
  static getItemStorageKey(itemKey, collectionKey) {
    if (itemKey == null || collectionKey == null)
      LogCodeManager.log("1A:02:00");
    if (itemKey == null)
      itemKey = "unknown";
    if (collectionKey == null)
      collectionKey = "unknown";
    return this.storageItemKeyPattern.replace("${collectionKey}", collectionKey.toString()).replace("${itemKey}", itemKey.toString());
  }
  static getGroupStorageKey(groupKey, collectionKey) {
    if (groupKey == null || collectionKey == null)
      LogCodeManager.log("1A:02:01");
    if (groupKey == null)
      groupKey = "unknown";
    if (collectionKey == null)
      collectionKey = "unknown";
    return this.storageGroupKeyPattern.replace("${collectionKey}", collectionKey.toString()).replace("${groupKey}", groupKey.toString());
  }
};
let CollectionPersistent = _CollectionPersistent;
CollectionPersistent.defaultGroupSideEffectKey = "rebuildGroupStorageValue";
CollectionPersistent.storageItemKeyPattern = "_${collectionKey}_item_${itemKey}";
CollectionPersistent.storageGroupKeyPattern = "_${collectionKey}_group_${groupKey}";

export { CollectionPersistent };
